/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include <sys/types.h>
#include "qwsvdef.h"

#ifdef NeXT
#include <libc.h>
#endif

#if defined(__linux__) || defined(sun)
#include <sys/stat.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>
#else
#include <sys/dir.h>
#endif

// mem�ria compartilhada
#include <sys/ipc.h>
#include <sys/shm.h>

// sem�foro
#include <sys/sem.h>

// clone
#include <syscall.h>

// wait
#include <sys/wait.h>

// affinity
#include <sched.h>
#include <linux/sched.h>

// kill
#include <signal.h>

cvar_t	sys_nostdout 		    = {"sys_nostdout",		    "0"};
cvar_t	sys_extrasleep 		    = {"sys_extrasleep",	    "0"};
cvar_t	par_debugmsgs		    = {"par_debugmsgs",		    "0"};
cvar_t	par_benchmark		    = {"par_benchmark",		    "1"};
cvar_t  par_pmovesize           = {"par_pmovesize",         "256"};
cvar_t  par_numprocs            = {"par_numprocs",          "1"};
cvar_t  par_edictfix            = {"par_edictfix",          "1"};
cvar_t  par_forceretouch        = {"par_forceretouch",      "1"};
cvar_t  par_sortingalgo         = {"par_sortingalgo",       "0"};
cvar_t  par_forceaffinity       = {"par_forceaffinity",     "1"};
cvar_t  par_deferredresponse    = {"par_deferredresponse",  "0"};
cvar_t  par_loadbalance         = {"par_loadbalance",       "1"};
cvar_t	par_clienttarget		= {"par_clienttarget",		"128"};
cvar_t  par_frametarget         = {"par_frametarget",       "5000"};
cvar_t	par_groupinfo			= {"par_groupinfo",			"0"};
cvar_t	par_sleeptime			= {"par_sleeptime",			"10"};
cvar_t	par_imbalance			= {"par_imbalance",			"1"};
cvar_t  par_huffman_trans       = {"par_huffman_trans",     "0"};
cvar_t	par_sendproc			= {"par_sendproc",			"0"};
cvar_t	par_packetsort			= {"par_packetsort",		"0"};

qboolean	stdin_ready;

statistics_t *stats;

// ==========================================================================

void*   Sys_AllocateSharedMemorySegment ( int size );
void    Sys_AllocateSharedMemory        ( void );
void    Sys_DestroySharedMemory         ( void );
void    Sys_ServerAffinity              ( void );
void    Sys_ChildAffinity               ( pid_t pid, int cpu );
pid_t   Sys_Fork                        ( int cpu );
void    Sys_Wait                        ( void );
void    Sys_ExitChild                   ( void );
void    Sys_SettingsRecap               ( void );
void    Sys_BenchmarkReport             ( void );

/*
===============================================================================

				REQUIRED SYS FUNCTIONS

===============================================================================
*/

/*
============
Sys_FileTime

returns -1 if not present
============
*/
int	Sys_FileTime (char *path)
{
	struct	stat	buf;
	
	if (stat (path,&buf) == -1)
		return -1;
	
	return buf.st_mtime;
}


/*
============
Sys_mkdir

============
*/
void Sys_mkdir (char *path)
{
	if (mkdir (path, 0777) != -1)
		return;
	if (errno != EEXIST)
		Sys_Error ("mkdir %s: %s",path, strerror(errno)); 
}


/*
================
Sys_DoubleTime
================
*/
double Sys_DoubleTime (void)
{
	struct timeval tp;
	struct timezone tzp;
	static int		secbase;

	gettimeofday(&tp, &tzp);
	
	if (!secbase)
	{
		secbase = tp.tv_sec;
		return tp.tv_usec/1000000.0;
	}
	
	return (tp.tv_sec - secbase) + tp.tv_usec/1000000.0;
}

/*
================
Sys_Error
================
*/
void Sys_Error (char *error, ...)
{
	va_list		argptr;
	char		string[1024];
	
	va_start (argptr,error);
	vsprintf (string,error,argptr);
	va_end (argptr);
	printf ("Fatal error: %s\n",string);

	Sys_DestroySharedMemory();
	
	exit (1);
}

/*
================
Sys_Printf
================
*/
void Sys_Printf (char *fmt, ...)
{
	va_list		argptr;
	static char		text[2048];
	unsigned char		*p;

	va_start (argptr,fmt);
	vsprintf (text,fmt,argptr);
	va_end (argptr);

	if (strlen(text) > sizeof(text))
		Sys_Error("memory overwrite in Sys_Printf");

    if (sys_nostdout.value)
        return;

	for (p = (unsigned char *)text; *p; p++) {
		*p &= 0x7f;
		if ((*p > 128 || *p < 32) && *p != 10 && *p != 13 && *p != 9)
			printf("[%02x]", *p);
		else
			putc(*p, stdout);
	}
	fflush(stdout);
}


/*
================
Sys_Quit
================
*/
void Sys_Quit (void)
{
	if( par_benchmark.value )
        Sys_BenchmarkReport();
    
    Sys_SettingsRecap();
    Sys_DestroySharedMemory();
    exit(0);
}

static int do_stdin = 1;

/*
================
Sys_ConsoleInput

Checks for a complete line of text typed in at the console, then forwards
it to the host command processor
================
*/
char *Sys_ConsoleInput (void)
{
	static char	text[256];
	int		len;

	if (!stdin_ready || !do_stdin)
		return NULL;		// the select didn't say it was ready
	stdin_ready = false;

	len = read (0, text, sizeof(text));
	if (len == 0) {
		// end of file
		do_stdin = 0;
		return NULL;
	}
	if (len < 1)
		return NULL;
	text[len-1] = 0;	// rip off the /n and terminate
	
	return text;
}

/*
=============
Sys_Init

Quake calls this so the system can register variables before host_hunklevel
is marked
=============
*/
void Sys_Init (void)
{
	struct sigaction act;

	Cvar_RegisterVariable (&sys_nostdout);
	Cvar_RegisterVariable (&sys_extrasleep);
	Cvar_RegisterVariable (&par_debugmsgs);
	Cvar_RegisterVariable (&par_benchmark);
    Cvar_RegisterVariable (&par_pmovesize);
    Cvar_RegisterVariable (&par_numprocs);  
    Cvar_RegisterVariable (&par_edictfix);
    Cvar_RegisterVariable (&par_forceretouch);
    Cvar_RegisterVariable (&par_sortingalgo);
    Cvar_RegisterVariable (&par_forceaffinity);
    Cvar_RegisterVariable (&par_deferredresponse);
    Cvar_RegisterVariable (&par_loadbalance);
    Cvar_RegisterVariable (&par_clienttarget);
    Cvar_RegisterVariable (&par_frametarget);
    Cvar_RegisterVariable (&par_groupinfo);
    Cvar_RegisterVariable (&par_sleeptime);
    Cvar_RegisterVariable (&par_imbalance);
    Cvar_RegisterVariable (&par_huffman_trans);
    Cvar_RegisterVariable (&par_sendproc);
    Cvar_RegisterVariable (&par_packetsort);

	Sys_AllocateSharedMemory();
	Sys_ServerAffinity();

	// pai n�o precisa dar wait no filho
	act.sa_handler = SIG_IGN;
	act.sa_flags = SA_NOCLDWAIT;
	sigaction (SIGCHLD, &act, NULL);
}

/*
=============
main
=============
*/
void main(int argc, char *argv[])
{
	double			time, oldtime, newtime;
	quakeparms_t	parms;
	fd_set	fdset;
	extern	int		net_socket;
    struct timeval timeout;
	int j;
    static int do_once = 0;

	memset (&parms, 0, sizeof(parms));

	COM_InitArgv (argc, argv);	
	parms.argc = com_argc;
	parms.argv = com_argv;

	parms.memsize = 16*1024*1024;

	j = COM_CheckParm("-mem");
	if (j)
		parms.memsize = (int) (Q_atof(com_argv[j+1]) * 1024 * 1024);
	if ((parms.membase = malloc (parms.memsize)) == NULL)
		Sys_Error("Can't allocate %ld\n", parms.memsize);

	parms.basedir = ".";

    svs.benchmark_active = 0;

/*
	if (Sys_FileTime ("id1/pak0.pak") != -1)
	else
		parms.basedir = "/raid/quake/v2";
*/

	SV_Init (&parms);

// run one frame immediately for first heartbeat
	SV_Frame (0.1);		

#ifdef DEPURAR
	setvbuf(stdout, (char *)NULL, _IONBF, 0);
#endif

    //
    // main loop
    //
    //if( par_benchmark.value ) {
	    stats->frames = 0;
	    stats->total_wait_time = 0.0;
	    svs.num_clients = 0;

	    for(j=0; j<NPROCS; j++) {
	        stats->packets[j] =
	        stats->responses[j] =
	        stats->weight[j] =
	        stats->edict_propagations[j] = 0;
	    }
    //}

	oldtime = Sys_DoubleTime () - 0.1;
	printf("Entering main loop!\n");
	while (1)
	{
	// select on the net socket and stdin
	// the only reason we have a timeout at all is so that if the last
	// connected client times out, the message would not otherwise
	// be printed until the next event.
		FD_ZERO(&fdset);
		if (do_stdin)
			FD_SET(0, &fdset);
		FD_SET(net_socket, &fdset);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		if (select (net_socket+1, &fdset, NULL, NULL, &timeout) == -1)
			continue;
		stdin_ready = FD_ISSET(0, &fdset);

	// find time passed since last cycle
		newtime = Sys_DoubleTime ();
		time = newtime - oldtime;
		oldtime = newtime;

        if( par_benchmark.value && svs.num_clients >= par_clienttarget.value && !do_once ) {
            svs.benchmark_active = 1;
            stats->start_time = Sys_DoubleTime();
            do_once = 1;
            printf("B E G I N  B E N C H M A R K\n");
        }

        if( par_benchmark.value && svs.benchmark_active ) {
        	++stats->frames;
        	if( par_benchmark.value == 1 ) {
                double finish = Sys_DoubleTime();           
                if( finish - stats->start_time >= 180 ) {
        		    printf("Benchmark complete after %g seconds\n",finish - stats->start_time);
        		    Sys_Quit();
        	    }
            } else if( par_benchmark.value == 2 ) {
                if( stats->frames == (int)par_frametarget.value ) {
                    printf("Benchmark complete after %i frames\n",stats->frames);
        		    Sys_Quit();
                }
            }
        }

		SV_Frame (time);		
		
		// extrasleep is just a way to generate a fucked up connection on purpose
		if (sys_extrasleep.value)
			usleep (sys_extrasleep.value);
	}	
}

// =============================================================================
//
// The functions below this section are part of the parallel server.
//
// =============================================================================

#define mask_len (sizeof(unsigned int))
#define MAX_IDS 10

int shmids[MAX_IDS];
void* shm_ptrs[MAX_IDS];
int num_ids = -1;
int sem_wait;

void Sys_SettingsRecap( void )
{    
    printf("\nS E T T I N G S  R E C A P\n\n");

    printf("par_debugmsgs: \t\t\t\t%i\n",(int)par_debugmsgs.value);
    printf("par_benchmark: \t\t\t\t%i\n",(int)par_benchmark.value);
    printf("par_pmovesize: \t\t\t\t%i\n",(int)par_pmovesize.value);
    printf("par_numprocs: \t\t\t\t%i\n",(int)par_numprocs.value);
    printf("par_edictfix: \t\t\t\t%i\n",(int)par_edictfix.value);
    printf("par_forceretouch: \t\t\t%i\n",(int)par_forceretouch.value);
    printf("par_sortingalgo: \t\t\t%i\n",(int)par_sortingalgo.value);
    printf("par_forceaffinity: \t\t\t%i\n",(int)par_forceaffinity.value);
    printf("par_deferredresponse: \t\t\t%i\n",(int)par_deferredresponse.value);
    printf("par_loadbalance: \t\t\t%i\n",(int)par_loadbalance.value);
    printf("par_clienttarget: \t\t\t%i\n",(int)par_clienttarget.value);
    printf("par_frametarget: \t\t\t%i\n",(int)par_frametarget.value);
    printf("par_sleeptime: \t\t\t\t%i\n",(int)par_sleeptime.value);
    printf("par_imbalance: \t\t\t\t%i\n",(int)par_imbalance.value);
    printf("par_huffman_trans: \t\t\t%i\n",(int)par_huffman_trans.value);
    printf("\n");
}

void Sys_BenchmarkReport( void )
{
    double total_time;
    unsigned int packets = 0;
    unsigned int responses = 0;
    int i, j;

    total_time = (Sys_DoubleTime() - stats->start_time);

    printf("\nB E N C H M A R K  R E P O R T\n\n");

    printf("Total running time: \t\t\t%g\n",total_time);
    printf("Total number of frames: \t\t%u\n",stats->frames);
    printf("Frames per second: \t\t\t%g\n\n",stats->frames / total_time);

    printf("Total number of groups: \t\t%u\n",stats->total_groups);
    printf("Total group size: \t\t\t%u\n",stats->tam_grupos);
    printf("Average group size: \t\t\t%g\n",((double)stats->tam_grupos) / stats->total_groups);
    printf("Groups per frame: \t\t\t%g\n",((double)stats->total_groups) / stats->frames);
    printf("Groups per second: \t\t\t%g\n\n",((double)stats->total_groups) / total_time);

    printf("Requests per processor:\t\t\t");
    for(i=0; i<NPROCS; i++) {
        packets += stats->packets[i];
        printf("%u ", stats->packets[i]);
    }
    printf("\nTotal requests received: \t\t%u\n",packets);
    printf("Requests per frame: \t\t\t%g\n",((double)packets) / stats->frames);
    printf("Requests per second: \t\t\t%g\n\n",((double)packets) / total_time);

    printf("Responses per process: \t\t\t");
    for(i=0; i<NPROCS; i++) {
        responses += stats->responses[i];
        printf("%u ", stats->responses[i]);
    }
    printf("\nTotal number of replies sent: \t\t%u\n",responses);
    printf("Responses per frame: \t\t\t%g\n",((double)responses) / stats->frames);
    printf("Responses per second: \t\t\t%g\n\n",((double)responses) / total_time);
    
    printf("Average intra-frame wait: \t\t%g\n",stats->total_wait_time / stats->frames);
    printf("Total wait-time: \t\t\t%g\n",stats->total_wait_time);
    printf("Wait percentage: \t\t\t%g%\n\n",(stats->total_wait_time / total_time) * 100.0);

    printf("Workloads: \t\t\t\t");
    for(i=0; i<NPROCS; i++)
        printf("%u ", stats->weight[i]);
    printf("\n");

    printf("Propagations: \t\t\t\t");
    for(i=0; i<NPROCS; i++)
    	printf("%u ", stats->edict_propagations[i]);
    printf("\n");

    for( i = 0; i < 180; ++i ) {
    	if( taskSamples[i][0] == 0 )
    		break;

    	for( j = 0; j < MAX_EDICTS; ++j ) {
    		if( taskSamples[i][j] > 0 ) {
    			printf("%i,",taskSamples[i][j]);
    		}
    	}

    	printf("\n");
    }

}

// Sys_AllocateSharedMemorySegment.
// Allocates a segment of shared memory of specified size.
// Returns the allocated shared memory segment.
void* Sys_AllocateSharedMemorySegment( int size )
{
	void *memory;
    char msg_erro[30];
    int shmid;

    ++num_ids;
    shmid = shmget(IPC_PRIVATE, size, IPC_CREAT|IPC_EXCL|0660);
    shmids[num_ids] = shmid;

    if(shmid == -1) {
        sprintf(msg_erro, "shmget / id = %d", num_ids);
        perror(msg_erro);
        exit(1);
    }

    memory = shmat(shmid, NULL, 0);
    if(memory == NULL) {
    	sprintf(msg_erro, "shmat / id = %d", num_ids);
    	perror(msg_erro);
    	exit(1);
    }

    shm_ptrs[num_ids] = memory;

    Con_Printf("Allocated shared memory: size (%i), id (%i), ptr (0x%X).\n",
    		size,shmid,memory);

    return memory;
}


// Aloca mem�ria compartilhada para o vetor svs.clients,
// que cont�m o estado dos clientes
// Sys_AllocateSharedMemory.
//
// Returns nothing.
void Sys_AllocateSharedMemory( void ) 
{
	Con_Printf("Allocating shared memory pool.\n");

    svs.shared_state = (entvars_t*)Sys_AllocateSharedMemorySegment(sizeof(entvars_t) * MAX_CLIENTS);
    svs.clients = (client_t*)Sys_AllocateSharedMemorySegment(sizeof(client_t) * MAX_CLIENTS);
    svs.edict_baselines = (entity_state_t*)Sys_AllocateSharedMemorySegment(sizeof(entity_state_t) * MAX_CLIENTS);
    svs.edict_insertions = (edict_t*)Sys_AllocateSharedMemorySegment((888 * 16) * (NPROCS+1));
    stats = (statistics_t*)Sys_AllocateSharedMemorySegment(sizeof(statistics_t));

    // sem_wait
    sem_wait = semget(IPC_PRIVATE, 1, IPC_CREAT | 0660);
    semctl(sem_wait, 0, SETVAL, 0);
}

void Sys_DestroySharedMemory( void ) 
{
    int i;
    char msg_erro[30];

    Con_Printf("Deallocating shared memory pool.\n");

    for( i = 0; i <= num_ids; i++ ) {
    	if( shmdt(shm_ptrs[i]) == -1 ) {
    		sprintf(msg_erro,"shmdt / id = %d",num_ids);
    		perror(msg_erro);
    	}

    	shmctl(shmids[i], IPC_RMID, NULL);
		Con_Printf("Removing shared memory: id (%i).\n",shmids[i]);
    }

	semctl(sem_wait, 0, IPC_RMID);
}

// Sys_ServerAffinity.
// Sets the CPU affinity of the server (coordinator) process.
// Returns nothing.
void Sys_ServerAffinity( void )
{
    cpu_set_t mask;
    CPU_ZERO(&mask);
    CPU_SET(0,&mask);
    unsigned long proc1 = 1UL;

    if( par_forceaffinity.value == 1 ) {
		if( sched_setaffinity(getpid(), mask_len, &proc1) == -1 ) {
			perror("server_affinity");
			exit(1);
		}
    } else if( par_forceaffinity.value == 2 ) {
    	if( sched_setaffinity(getpid(), mask_len, &mask) == -1 ) {
    		perror("server_affinity");
    		exit(1);
    	}
    }
}

// Sys_ChildAffinity.
// Sets the CPU affinity of the child processes.
// Return nothing.
void Sys_ChildAffinity( pid_t pid, int cpu ) 
{
    cpu_set_t mask;
    CPU_ZERO(&mask);
    CPU_SET(cpu,&mask);
    unsigned long proc2 = 2UL;

    if( par_forceaffinity.value == 1 ) {
    	if( sched_setaffinity(pid, mask_len, &proc2) == -1 ) {
    		perror("child_affinity");
    		exit(1);
    	}
    } else if( par_forceaffinity.value == 2 ) {
    	if( sched_setaffinity(pid, mask_len, &mask) == -1 ) {
    		perror("child_affinity");
    		exit(1);
    	}
    }
}

// Sys_Fork.
// Creates a child process to use in parallel server execution.
// Returns the PID of the child process. Returns 0 if called from a child process.
pid_t Sys_Fork( int cpu )
{
    pid_t pid_child;

    pid_child = syscall(__NR_clone, CLONE_STOPPED|CLONE_CHILD_CLEARTID|CLONE_CHILD_SETTID|SIGCHLD, NULL);

	if(pid_child == -1) {
		perror("clone");
		exit(1);
	} else if(pid_child != 0) {
		// Child is required to run on processor 2.
        if( par_forceaffinity.value )		
            Sys_ChildAffinity(pid_child,cpu);

		// inicia execu��o do filho
		kill(pid_child, SIGCONT);
	}

	return pid_child;
}

// Sys_Wait.
//
// Returns nothing.
void Sys_Wait( void )
{
	struct sembuf smb;

	smb.sem_num =  0;
	smb.sem_flg =  0;
	smb.sem_op  = -1;

	semop(sem_wait, &smb, 1);

	/*
	 * if( waitpid(WAIT_ANY, NULL, __WALL) == -1 ) {
	 * perror("sys_wait");
	 * exit(1);
	 * }
	 */
}

// Sys_ExitChild.
//
// Returns nothing.CLONE_CHILD_CLEARTID|CLONE_CHILD_SETTID|SIGCHLD
void Sys_ExitChild( void )
{
    struct sembuf smb;

	smb.sem_num = 0;
	smb.sem_flg = 0;
	smb.sem_op  = 1;

	semop(sem_wait, &smb, 1);

	exit(EXIT_SUCCESS);
}
